package com.ringlord.antbuilder;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Window;
import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.Insets;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;

import java.io.File;
import java.io.Writer;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.net.URL;

import java.util.StringTokenizer;
import java.util.Map;
import java.util.HashMap;

import java.util.prefs.Preferences;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JTextField;
import javax.swing.JComboBox;
import javax.swing.JPasswordField;
import javax.swing.JTextArea;
import javax.swing.JCheckBox;
import javax.swing.JTabbedPane;
import javax.swing.JScrollPane;
import javax.swing.JRadioButton;
import javax.swing.JFileChooser;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.WindowConstants;
import javax.swing.SwingUtilities;


/**
 * A tool to construct an Ant "build.xml" file based on a variety of
 * desired features and choices for the project.
 *
 * @author K. Udo Schuermann
 **/
class Main
  extends JFrame
  implements Runnable,
             ActionListener,
             WindowListener
{
  public static final String VERSION = "1.23";
  public static final String VERDATE = "13-Jul-2013";
  public static final String EOLN = System.getProperty("line.separator");
  public static final String[] SOURCE_GENERATION =
  {
    "None",
    "Swing (Java 1.5+)",
    "JavaFX (Java 1.8+)"
  };

  public static void main( final String[] args )
  {
    final Main main = new Main( args );
    SwingUtilities.invokeLater( main );
    System.err.println( "AntBuilder "+VERSION+" ("+VERDATE+")" );
    System.err.println( "Copyright (C) 2009-2013 Udo Schuermann" );
    System.err.println( "All rights reserved." );
    System.err.println();
    System.err.println( "Home page: http://ringlord.com/antbuilder.html" );
    System.err.println();
    System.err.println( "AntBuilder is licensed under the GNU General Public License v3+" );
    System.err.println( "A copy of the the license (LICENSE.TXT) is available inside the" );
    System.err.println( "antbuilder.jar file, along with AntBuilder's source code and the" );
    System.err.println( "means to reproduce the software (Note: .jar files are .zip files)" );
  }

  Main( final String[] args )
  {
    super( "AntBuilder "+VERSION );
    this.args = args;
  }

  public void run()
  {
    final Form optMain = new Form();
    final Form optMisc = new Form();
    final Form optJAXB = new Form();
    final Form optPath = new Form();

    final String curJavaVersion = System.getProperty( "java.specification.version" );

    this.projectName      = new JTextField();
    this.projectWAR       = new JCheckBox("Build a WAR instead of a JAR");
    this.projectFile      = new JTextField();
    this.projectMain      = new JTextField();
    this.projectSVer      = new JTextField( curJavaVersion );
    this.projectTVer      = new JTextField( curJavaVersion );
    this.projectRepackageLibs = new JRadioButton("Merge libraries into executable");
    this.projectReferenceLibs = new JRadioButton("Reference libraries in Class-Path");
    this.projectImg       = new JCheckBox("Include img/* (image files) in executable");
    this.projectEtc       = new JCheckBox("Include etc/* (miscellaneous files) in executable" );
    this.projectHelp      = new JCheckBox("Include help/* (help files) in executable" );
    this.projectSrc       = new JCheckBox("Include src/* (source code) in executable");
    this.projectSignX     = new JCheckBox("Sign JAR (required for Applets & Webstart)");
    this.projectSignL     = new JCheckBox("Sign libraries");
    this.projectSignInfoFile = new JTextField();
    this.projectGit       = new JRadioButton("Build project.log file from Git log");
    this.projectSvn       = new JRadioButton("Build project.log file from SVN log");
    this.projectSkeleton  = new JComboBox<String>( SOURCE_GENERATION );
    this.projectJAXB      = new JCheckBox("Build JAXB-related files");
    this.projectChkSum    = new JCheckBox("Create MD5, SHA1, and SHA-256 Checksum files");
    //
    this.pathSrc          = new JTextField();
    this.pathLib          = new JTextField();
    this.pathImg          = new JTextField();
    this.pathHelp         = new JTextField();
    this.pathEtc          = new JTextField();
    this.pathBuild        = new JTextField();

    this.noLibraries = new JRadioButton("No libraries", true);
    final ButtonGroup references = new ButtonGroup();
    references.add( this.projectRepackageLibs );
    references.add( this.projectReferenceLibs );
    references.add( this.noLibraries );

    this.noLogs = new JRadioButton("No project.log file", true);
    final ButtonGroup logs = new ButtonGroup();
    logs.add( this.projectGit );
    logs.add( this.projectSvn );
    logs.add( this.noLogs );

    projectName.setText( "Example Project" );
    projectFile.setText( "example" );
    projectMain.setText( "net.example.project.Main" );
    projectImg.setSelected( false );
    projectSrc.setSelected( false );
    pathSrc.setText( "src" );
    pathLib.setText( "lib" );
    pathImg.setText( "img" );
    pathHelp.setText( "help" );
    pathEtc.setText( "etc" );
    pathBuild.setText( ".build" );

    final JFrame myFrame = this; // for the action listener
    final ClassLoader cl = getClass().getClassLoader();

    final java.net.URL folderIcon = cl.getResource("img/folder.png");
    this.projectSignInfoPick =
      new JButton( new javax.swing.ImageIcon(cl.getResource("img/folder.png")) );
    projectSignInfoPick.setMargin( new java.awt.Insets(0,0,0,0) );
    projectSignInfoPick.addActionListener( new ActionListener()
      {
        public void actionPerformed( final ActionEvent e )
        {
          int pick = d.showOpenDialog( myFrame );
          if( pick == JFileChooser.APPROVE_OPTION )
            {
              projectSignInfoFile.setText( d.getSelectedFile().getName() );
            }
        }
        private final JFileChooser d = new JFileChooser()
          {
            {
              setDialogTitle( "Project Signing Properties File" );
              setFileHidingEnabled( false );
              final File cwd = new File( System.getProperty("user.dir") );
              setCurrentDirectory( cwd.getAbsoluteFile().getParentFile() );
            }
          };
      } );

    final JPanel signingPropertiesPanel = new JPanel( new BorderLayout() );
    signingPropertiesPanel.add( BorderLayout.CENTER, projectSignInfoFile );
    signingPropertiesPanel.add( BorderLayout.EAST,   projectSignInfoPick );

    optMain.addRow( mkLabel("Project name:",'n',projectName),        projectName );
    optMain.addRow( mkLabel("Executable:",'x',projectFile),          projectFile );
    optMain.addRow( mkLabel("Main Class:",'M',projectMain),          projectMain );
    optMain.addRow( mkLabel("Java Source Version:",'S',projectSVer), projectSVer );
    optMain.addRow( mkLabel("Java Target Version:",'T',projectTVer), projectTVer );
    optMain.addRow( mkLabel("Create skeleton:",'k',projectSkeleton), projectSkeleton );
    optMain.addRow( mkLabel("Images:",'i',projectImg),               projectImg );
    optMain.addRow( mkLabel("Miscellaneous:",'l',projectEtc),        projectEtc );
    optMain.addRow( mkLabel("Help files:",'H',projectHelp),          projectHelp );
    optMain.addRow( mkLabel("Sources:",'u',projectSrc),              projectSrc );
    optMain.addRow( mkLabel("Package:",'g',projectWAR),              projectWAR );

    optMisc.addRow( mkLabel("Libraries (.jar/.zip):",'L',projectRepackageLibs), projectRepackageLibs );
    optMisc.addRow( new JLabel(""),                                  projectReferenceLibs );
    optMisc.addRow( new JLabel(""),                                  noLibraries );
    optMisc.addRow( mkLabel("Sign JAR:",'S',projectSignX),           projectSignX );
    optMisc.addRow( mkLabel("Re-sign Libraries:",'R',projectSignL),  projectSignL );
    optMisc.addRow( mkLabel("Signing Properties:",'g',projectSignInfoFile), signingPropertiesPanel );
    optMisc.addRow( mkLabel("Checksum Files:",'h',projectChkSum),    projectChkSum );
    optMisc.addRow( mkLabel("Project log source",'o',projectGit),    projectGit );
    optMisc.addRow( new JLabel(""),                                  projectSvn );
    optMisc.addRow( new JLabel(""),                                  noLogs );

    optJAXB.addRow( mkLabel("Include JAXB:",'X',projectJAXB),        projectJAXB );

    optPath.addRow( mkLabel("Source directory:",'S',pathSrc),        pathSrc );
    optPath.addRow( mkLabel("Library directory:",'L',pathLib),       pathLib );
    optPath.addRow( mkLabel("Images directory:",'I',pathImg),        pathImg );
    optPath.addRow( mkLabel("Help directory:",'H',pathHelp),         pathHelp );
    optPath.addRow( mkLabel("Miscellaneous directory:",'l',pathEtc), pathEtc );
    optPath.addRow( mkLabel("Build directory:",'B',pathBuild),       pathBuild );

    this.accept = new JButton( "Accept",
                               new javax.swing.ImageIcon(cl.getResource("img/accept.png")) );
    this.accept.setMnemonic('A');
    this.cancel = new JButton( "Cancel",
                               new javax.swing.ImageIcon(cl.getResource("img/cancel.png")) );
    this.cancel.setMnemonic('C');

    final JPanel cmd = new JPanel( new FlowLayout(FlowLayout.RIGHT) );
    cmd.add( this.accept );
    cmd.add( this.cancel );

    final JPanel mainPanel = new JPanel( new BorderLayout() );
    mainPanel.add( BorderLayout.NORTH, optMain );

    final JPanel miscPanel = new JPanel( new BorderLayout() );
    miscPanel.add( BorderLayout.NORTH, optMisc );

    final JPanel jaxbPanel = new JPanel( new BorderLayout() );
    jaxbPanel.add( BorderLayout.NORTH, optJAXB );

    final JPanel pathPanel = new JPanel( new BorderLayout() );
    pathPanel.add( BorderLayout.NORTH, optPath );

    final JTabbedPane tab = new JTabbedPane();
    tab.addTab( "Main",  mainPanel );
    tab.addTab( "Other", miscPanel );
    tab.addTab( "Paths", pathPanel );
    tab.addTab( "JAXB",  jaxbPanel );

    tab.setMnemonicAt( 0, 'M' );
    tab.setMnemonicAt( 1, 'O' );
    tab.setMnemonicAt( 2, 'P' );
    tab.setMnemonicAt( 3, 'J' );

    final JPanel p = new JPanel( new BorderLayout() );
    p.add( tab, BorderLayout.CENTER );
    p.add( cmd, BorderLayout.SOUTH );

    getContentPane().add( p );

    this.projectName.setToolTipText( "Name of your project in the build.xml" );

    this.projectFile.setToolTipText( "<html>"+
                                     "The name of the executable file to be<br>"+
                                     "generated. The filename extension is<br>"+
                                     "determined by the 'type' of executable<br>"+
                                     "(i.e. .jar or .war)" );

    this.projectWAR.setToolTipText( "<html>"+
                                    "If you don't know why you might want a<br>"+
                                    "WAR file, then stick to JAR files, instead." );

    this.projectMain.setToolTipText( "<html>"+
                                     "The fully qualified class name of your<br>"+
                                     "application's entry point. This becomes<br>"+
                                     "the value of the Main-Class property in<br>"+
                                     "JAR's manifest." );

    this.projectSrc.setToolTipText( "<html>"+
                                    "Free Software and Open Source Software<br>"+
                                    "may want to ship the source code directly<br>"+
                                    "in the executable.<br>"+
                                    "<br>"+
                                    "For proprietary software, avoid checking<br>"+
                                    "this option." );

    this.projectRepackageLibs.setToolTipText( "<html>"+
                                              "Transfers the contents of .jar and .zip files<br>"+
                                              "from the lib/ directory into your executable.<br>"+
                                              "<br>"+
                                              "This makes for a single executable without<br>"+
                                              "external dependencies." );

    this.projectReferenceLibs.setToolTipText( "<html>"+
                                              "References .jar and .zip files in the lib/<br>"+
                                              "directory using the Class-Path property<br>"+
                                              "in the JAR's manifest.<br>"+
                                              "<br>"+
                                              "These become external dependencies<br>"+
                                              "for your project. You need to ship these<br>"+
                                              "files along with your software, possibly<br>"+
                                              "by referencing them in a .jnlp file and<br>"+
                                              "serving them from your server." );
    this.noLibraries.setToolTipText( "No libraries are used explicitly." );

    this.projectSVer.setToolTipText( "<html>"+
                                     "The version of your source code. If you use<br>"+
                                     "generics then that will be at least 1.5; if<br>"+
                                     "you use try-with-resources, diamond operator,<br>"+
                                     "String in switch/case, etc. then you want to<br>"+
                                     "set this at least to 1.7<br>"+
                                     "<br>"+
                                     "If you use a later compiler than this version<br>"+
                                     "you may see the compiler complain thus:<br>"+
                                     "'bootstrap class path not set in conjuection<br>"+
                                     "with -source \u2026'; this means that your<br>"+
                                     "current libraries offer classes and methods<br>"+
                                     "that earlier sources may not have had yet,<br>"+
                                     "and the compiler cannot help you unless you<br>"+
                                     "provide a 'bootclasspath' argument to the<br>"+
                                     "Ant javac target with the full path to the<br>"+
                                     "older rt.jar file.<br>"+
                                     "<br>"+
                                     "Your target version must be at least as high<br>"+
                                     "as the source version." );

    this.projectTVer.setToolTipText( "<html>"+
                                     "To reach the most people, you want this to be as<br>"+
                                     "low as possible, but no lower than your source<br>"+
                                     "code version. Compiler/JVM performance reasons<br>"+
                                     "are probably the only reason to set this to a<br>"+
                                     "higher value.<br>"+
                                     "<br>"+
                                     "The version you select here will be the default,<br>"+
                                     "but you can arbitrarily target higher versions,<br>"+
                                     "too. If your target is '1.7' you can invoke Ant<br>"+
                                     "with a jar8 target, instead, to compile with a<br>"+
                                     "Java 8 compiler (provided that you have that)." );

    this.projectSignX.setToolTipText( "<html>"+
                                      "Digitally sign the JAR file. This is required<br>"+
                                      "for applications delivered via Java Webstart.<br>"+
                                      "<br>"+
                                      "Ensure that the signature information<br>"+
                                      "is available in the 'Signing property file'" );

    this.projectSignL.setToolTipText( "<html>"+
                                      "Digitally sign .jar files in lib/<br>"+
                                      "<br>"+
                                      "If you deliver your software with Java Web Start,<br>"+
                                      "externally referenced libraries also need to be<br>"+
                                      "digitally signed if your software needs access to<br>"+
                                      "local resources.<br>"+
                                      "<br>"+
                                      "Additionally, many (most? all?) implementations of<br>"+
                                      "Java Web Start require that the executable and<br>"+
                                      "the libraries are signed with the <em>same</em> key.<br>"+
                                      "<br>"+
                                      "Ensure that signature information is available in<br>"+
                                      "the 'Signing property file' (see below)" );

    this.projectSignInfoFile.setToolTipText( "<html>"+
                                             "This file must provide these properties:<br>"+
                                             "<ol>"+
                                             "<li>keystore-file"+
                                             "<li>keystore-pass"+
                                             "<li>key-alias"+
                                             "<li>key-pass</ol>"+
                                             "Do NOT keep this file with your project as<br>"+
                                             "the passwords are likely sensitive data!<br>"+
                                             "Example: <tt>../.keystore-info</tt>" );

    this.projectChkSum.setToolTipText( "<html>"+
                                       "Checksum files are used traditionally to<br>"+
                                       "verify the integrity of a transmitted file.<br>"+
                                       "Verification would go like this:<br>"+
                                       "<br>"+
                                       "<code>md5sum -c myproj.jar.md5</code><br>"+
                                       "<code>sha1sum -c myproj.jar.sha1</code><br>"+
                                       "<code>sha256sum -c myproj.jar.sha256</code>" );

    this.projectGit.setToolTipText( "<html>"+
                                    "The 'ant log' target generates a .log file thus:<br>"+
                                    "<code>git -M --name-status --graph --fullhistory</code><br>" );

    this.projectSvn.setToolTipText( "<html>"+
                                    "The 'ant log' target generates a .log file thus:<br>"+
                                    "<code>svn -v -r HEAD:1</code>" );
    this.noLogs.setToolTipText( "No 'log' target is generated" );

    this.projectSkeleton.setToolTipText( "<html>"+
                                         "Create skeleton source files in src/<br>"+
                                         "to get you started on your project.<br>"+
                                         "<br>"+
                                         "The code is nicely documented, and<br>"+
                                         "you can always delete the files again<br>"+
                                         "if you didn't want them. Your own<br>"+
                                         "files will never be altered.<br>"+
                                         "<br>"+
                                         "You must ensure that the source and<br>"+
                                         "target versions match the generated<br>"+
                                         "code, otherwise compilation will fail." );

    this.projectJAXB.setToolTipText( "<html>"+
                                     "This is here only because it was useful on one<br>"+
                                     "project. It's essentially unsupported but may<br>"+
                                     "be useful to some. Look at the <code>build.xml</code> to<br>"+
                                     "find out what all this option does." );

    this.accept.setToolTipText( "<html>"+
                                "Save settings and write<br>"+
                                "the _build.xml file." );
    this.cancel.setToolTipText( "<html>"+
                                "Forget all changes, write<br>"+
                                "nothing, and exit." );

    this.addWindowListener( this );
    this.setDefaultCloseOperation( WindowConstants.DO_NOTHING_ON_CLOSE );

    this.accept.addActionListener( this );
    this.cancel.addActionListener( this );

    this.projectRepackageLibs.addActionListener( new ActionListener()
      {
        public void actionPerformed( final ActionEvent e )
        {
          if( projectRepackageLibs.isSelected() )
            {
              projectReferenceLibs.setSelected( false );
            }
          projectSignL.setEnabled( false );
          if( projectSignX.isSelected() )
            {
              projectSignInfoFile.setEnabled( true );
              projectSignInfoPick.setEnabled( true );
            }
          else
            {
              projectSignInfoFile.setEnabled( false );
              projectSignInfoPick.setEnabled( false );
            }
        }
      } );
    this.projectReferenceLibs.addActionListener( new ActionListener()
      {
        public void actionPerformed( final ActionEvent e )
        {
          if( projectReferenceLibs.isSelected() )
            {
              projectRepackageLibs.setSelected( false );
              projectSignL.setEnabled( true );
            }
          else
            {
              projectSignL.setEnabled( false );
            }
          if( projectSignX.isSelected() ||
              projectSignL.isSelected() )
            {
              projectSignInfoFile.setEnabled( true );
              projectSignInfoPick.setEnabled( true );
            }
          else
            {
              projectSignInfoFile.setEnabled( false );
              projectSignInfoPick.setEnabled( false );
            }
        }
      } );
    this.noLibraries.addActionListener( new ActionListener()
      {
        public void actionPerformed( final ActionEvent e )
        {
          projectSignL.setEnabled( false );
          if( projectSignX.isSelected() )
            {
              projectSignInfoFile.setEnabled( true );
              projectSignInfoPick.setEnabled( true );
            }
          else
            {
              projectSignInfoFile.setEnabled( false );
              projectSignInfoPick.setEnabled( false );
            }
        }
      } );

    this.projectSignX.addActionListener( this );
    this.projectSignL.addActionListener( this );

    load();

    // Initial state of interdependent fields
    projectSignL.setEnabled( projectReferenceLibs.isSelected() );
    if( projectSignX.isSelected() ||
        (projectSignL.isSelected() &&
         projectReferenceLibs.isSelected()) )
      {
        projectSignInfoFile.setEnabled( true );
        projectSignInfoPick.setEnabled( true );
      }
    else
      {
        projectSignInfoFile.setEnabled( false );
        projectSignInfoPick.setEnabled( false );
      }

    pack();
    positionWindow( this,
                    prefs.get("window-main-xywh",
                              null) );
    setVisible( true );
  }

  private static JLabel mkLabel( final String text,
                                 final Character mnemonic,
                                 final JComponent labelFor )
  {
    final JLabel label = new JLabel( text, JLabel.TRAILING );
    if( mnemonic != null )
      {
        label.setDisplayedMnemonic( mnemonic );
      }
    if( labelFor != null )
      {
        label.setLabelFor( labelFor );
      }
    return label;
  }

  private static void positionWindow( final Window frame,
                                      final String xywh )
  {
    final Dimension screenDim = frame.getToolkit().getScreenSize();
    final Rectangle windowDim = frame.getBounds();

    if( xywh != null )
      {
        final StringTokenizer t = new StringTokenizer( xywh, ", \t" );
        double[] dims = new double[4];
        int n = 0;
        while( (n < 4) &&
               t.hasMoreTokens() )
          {
            try
              {
                final double d = Double.parseDouble( t.nextToken() );
                if( (d < 0.0d) ||
                    (d > 1.0d) )
                  {
                    // bad range of values (must be [0..1]): do not cause
                    // n++ to increment, so the first test outside the
                    // loop prevents actual use of these values.
                    break;
                  }
                dims[n++] = d;
              }
            catch( final Exception x )
              {
                // couldn't parse the double, so ignore the exported
                // dimensions (which are bogus).
                break;
              }
          }
        if( n >= 4 ) // got all four values?
          {
            windowDim.x = (int)((double)screenDim.width * dims[0]);
            windowDim.y = (int)((double)screenDim.height * dims[1]);
            windowDim.width = (int)((double)screenDim.width * dims[2]);
            windowDim.height = (int)((double)screenDim.height * dims[3]);
          }
      }
    else
      {
        // Position the window slightly above the horizontal center

        final int x = (int)(Math.abs(screenDim.width-windowDim.width)*0.5d);
        final int y = (int)(Math.abs(screenDim.height-windowDim.width)*0.3333d);
        final int w = Math.min(screenDim.width,windowDim.width);
        final int h = Math.min(screenDim.height,windowDim.height);

        windowDim.x = x;
        windowDim.y = y;
        windowDim.width = w;
        windowDim.height = h;
      }

    // Ensure that a window that was closed too tiny is at least a
    // minimum size
    if( windowDim.width < 160 )
      {
        windowDim.width = 160;
      }
    if( windowDim.height < 100 )
      {
        windowDim.height = 100;
      }

    frame.setBounds( windowDim );
  }
  public static String getXYWH( final Window frame )
  {
    final Dimension screenDim = frame.getToolkit().getScreenSize();
    final Rectangle windowDim = frame.getBounds();

    final double x = (double)windowDim.x / (double)screenDim.width;
    final double y = (double)windowDim.y / (double)screenDim.height;
    final double w = (double)windowDim.width / (double)screenDim.width;
    final double h = (double)windowDim.height / (double)screenDim.height;

    return x+","+y+","+w+","+h;
  }

  private void load()
  {
    BufferedReader f = null;
    try
      {
        f = new BufferedReader( new FileReader(".antbuildrc") );
        String s;
        while( (s = f.readLine()) != null )
          {
            if( (s.length() == 0) ||
                (s.charAt(0) == '#') )
              {
                continue;
              }
            final int eqPos = s.indexOf( "=" );
            if( eqPos < 0 )
              {
                continue;
              }
            final String field = s.substring( 0, eqPos ).trim();
            final String value = s.substring( eqPos+1 );

            if( field.equals( "project-name" ) )
              {
                projectName.setText( value );
              }
            else if( field.equals( "project-is-war" ) )
              {
                projectWAR.setSelected( "true".equals(value) );
              }
            else if( field.equals( "executable-file" ) )
              {
                projectFile.setText( value );
              }
            else if( field.equals( "main-class" ) )
              {
                projectMain.setText( value );
              }
            else if( field.equals( "default-source-version" ) )
              {
                projectSVer.setText( value );
              }
            else if( field.equals( "default-java-version" ) )
              {
                projectTVer.setText( value );
              }
            else if( field.equals( "repackage-libs" ) )
              {
                // deprecated in favor of 'libraries'
                projectRepackageLibs.setSelected( "true".equals(value) );
              }
            else if( field.equals( "include-lib" ) )
              {
                // deprecated in favor of 'libraries'
                projectReferenceLibs.setSelected( "true".equals(value) );
              }
            else if( field.equals( "libraries" ) )
              {
                if( "internalize".equals(value) )
                  {
                    projectRepackageLibs.setSelected( true );
                  }
                else if( "reference".equals(value) )
                  {
                    projectReferenceLibs.setSelected( true );
                  }
                else
                  {
                    noLibraries.setSelected( true );
                  }
              }
            else if( field.equals( "include-img" ) )
              {
                projectImg.setSelected( "true".equals(value) );
              }
            else if( field.equals( "include-src" ) )
              {
                projectSrc.setSelected( "true".equals(value) );
              }
            else if( field.equals( "include-etc" ) )
              {
                projectEtc.setSelected( "true".equals(value) );
              }
            else if( field.equals( "include-help" ) )
              {
                projectHelp.setSelected( "true".equals(value) );
              }
            else if( field.equals( "checksum-executable" ) )
              {
                projectChkSum.setSelected( "true".equals(value) );
              }
            else if( field.equals( "sign-executable" ) )
              {
                projectSignX.setSelected( "true".equals(value) );
              }
            else if( field.equals( "sign-properties-file" ) )
              {
                projectSignInfoFile.setText( value );
              }
            else if( field.equals( "sign-libraries" ) )
              {
                projectSignL.setSelected( "true".equals(value) );
              }
            else if( field.equals( "git-support" ) )
              {
                // deprecated in favor of log-support
                projectGit.setSelected( "true".equals(value) );
              }
            else if( field.equals( "svn-support" ) )
              {
                // deprecated in favor of log-support
                projectSvn.setSelected( "true".equals(value) );
              }
            else if( field.equals( "log-support" ) )
              {
                if( "git".equals(value) )
                  {
                    projectGit.setSelected( true );
                  }
                else if( "svn".equals(value) )
                  {
                    projectSvn.setSelected( true );
                  }
                else
                  {
                    noLogs.setSelected( true );
                  }
              }
            else if( field.equals( "create-skeleton" ) )
              {
                if( "true".equals(value) )
                  {
                    // "true" used to mean "Yes, generate Java sources"
                    projectSkeleton.setSelectedItem( SOURCE_GENERATION[1] );
                  }
                else if( "false".equals(value) )
                  {
                    // "false" used to mean "No, don't generate Java sources"
                    projectSkeleton.setSelectedItem( SOURCE_GENERATION[0] );
                  }
                else
                  {
                    // All future values are from a larger set
                    projectSkeleton.setSelectedItem( value );
                  }
              }
            else if( field.equals( "jaxb-support" ) )
              {
                projectJAXB.setSelected( "true".equals(value) );
              }
            else if( field.equals( "path-src" ) )
              {
                pathSrc.setText( value );
              }
            else if( field.equals( "path-lib" ) )
              {
                pathLib.setText( value );
              }
            else if( field.equals( "path-img" ) )
              {
                pathImg.setText( value );
              }
            else if( field.equals( "path-help" ) )
              {
                pathHelp.setText( value );
              }
            else if( field.equals( "path-etc" ) )
              {
                pathEtc.setText( value );
              }
            else if( field.equals( "path-build" ) )
              {
                pathBuild.setText( value );
              }
            else
              {
                System.err.println( "Warning: Unrecognized property "+
                                    "in file '.antbuildrc': "+field );
              }
          }
      }
    catch( FileNotFoundException x )
      {
        // Silently ignore a missing file
      }
    catch( IOException x )
      {
        System.err.println( "Warning: Failed to read file '.antbuildrc': "+
                            x.getMessage() );
      }
    finally
      {
        if( f != null )
          {
            try
              {
                f.close();
              }
            catch( IOException x )
              {
                // it's okay
              }
          }
      }
  }

  private void save()
  {
    Writer f = null;
    try
      {
        f = new FileWriter( ".antbuildrc" );
        f.write( "# This file stores the choices you made in AntBuilder."+EOLN+
                 "# It simplifies rebuilding Ant's build.xml file if you"+EOLN+
                 "# wish to adjust your choices at a later time but it's"+EOLN+
                 "# not critical to keep around if you really don't want"+EOLN+
                 "# to keep it; be careful of modifying it by hand."+EOLN+
                 EOLN );

        f.write( "project-name="+projectName.getText()+EOLN );
        f.write( "project-is-war="+projectWAR.isSelected()+EOLN );
        f.write( "executable-file="+projectFile.getText()+EOLN );
        f.write( "main-class="+projectMain.getText()+EOLN );
        f.write( "default-source-version="+projectSVer.getText()+EOLN );
        f.write( "default-java-version="+projectTVer.getText()+EOLN );
        f.write( "libraries="+(projectRepackageLibs.isSelected()
                               ? "internalize"
                               : (projectReferenceLibs.isSelected()
                                  ? "reference"
                                  : ""))+EOLN );
        f.write( "include-img="+projectImg.isSelected()+EOLN );
        f.write( "include-src="+projectSrc.isSelected()+EOLN );
        f.write( "include-etc="+projectEtc.isSelected()+EOLN );
        f.write( "include-help="+projectHelp.isSelected()+EOLN );
        f.write( "checksum-executable="+projectChkSum.isSelected()+EOLN );
        f.write( "log-support="+(projectGit.isSelected()
                                 ? "git"
                                 : (projectSvn.isSelected()
                                    ? "svn"
                                    : ""))+EOLN );
        f.write( "create-skeleton="+projectSkeleton.getSelectedItem()+EOLN );
        f.write( "sign-executable="+projectSignX.isSelected()+EOLN );
        f.write( "sign-libraries="+projectSignL.isSelected()+EOLN );
        f.write( "sign-properties-file="+projectSignInfoFile.getText()+EOLN );
        f.write( "jaxb-support="+projectJAXB.isSelected()+EOLN );
        f.write( "path-src="+pathSrc.getText()+EOLN );
        f.write( "path-lib="+pathLib.getText()+EOLN );
        f.write( "path-img="+pathImg.getText()+EOLN );
        f.write( "path-help="+pathHelp.getText()+EOLN );
        f.write( "path-etc="+pathEtc.getText()+EOLN );
        f.write( "path-build="+pathBuild.getText()+EOLN );

        f.write( EOLN+
                 "#eof"+EOLN );

        f.flush();
      }
    catch( IOException x )
      {
        System.err.println( "Warning: Unable to save current choices "+
                            "to file '.antbuildrc': "+x.getMessage() );
      }
    finally
      {
        if( f != null )
          {
            try
              {
                f.close();
              }
            catch( IOException x )
              {
                // it's okay...
              }
          }
      }
  }

  private void buildXML()
    throws IOException
  {
    String signInfoFile = projectSignInfoFile.getText();
    BuildXMLWriter.buildXML( projectName.getText(),
                             projectWAR.isSelected(),
                             projectFile.getText(),
                             projectMain.getText(),
                             projectSVer.getText(),
                             projectTVer.getText(),
                             (projectRepackageLibs.isSelected()
                              ? "internalize"
                              : (projectReferenceLibs.isSelected()
                                 ? "reference"
                                 : "") ),
                             projectImg.isSelected(),
                             projectSrc.isSelected(),
                             projectEtc.isSelected(),
                             projectHelp.isSelected(),
                             projectChkSum.isSelected(),
                             projectSignX.isSelected() ,
                             projectSignL.isSelected() && projectSignL.isEnabled(),
                             translate(signInfoFile),
                             (projectGit.isSelected()
                              ? "git"
                              : (projectSvn.isSelected()
                                 ? "svn"
                                 : "")),
                             (String)projectSkeleton.getSelectedItem(),
                             projectJAXB.isSelected(),
                             pathSrc.getText(),
                             pathLib.getText(),
                             pathImg.getText(),
                             pathHelp.getText(),
                             pathEtc.getText(),
                             pathBuild.getText() );
  }

  /**
   * Translates a few common Unix-shell property names and constructs
   * into Java equivalents.
   **/
  private String translate( String value )
  {
    if( value.startsWith("~/") )
      {
        value = "${user.home}"+value.substring(1);
      }
    final int homeIndex = value.indexOf("${HOME}");
    if( homeIndex >= 0 )
      {
        value = value.substring(0,homeIndex) + "${user.home}" + value.substring(homeIndex+7);
      }
    return value;
  }

  private void shutdown()
  {
    setVisible( false );
    prefs.put( "window-main-xywh",
               getXYWH(this) );
  }

  // ======================================================================
  // ActionListener
  // ======================================================================
  public void actionPerformed( final ActionEvent e )
  {
    final Object o = e.getSource();
    if( o == accept )
      {
        setVisible( false );
        try
          {
            buildXML();
            System.err.println( SUCCESS_MESSAGE );
            save();
            System.exit( 0 );
          }
        catch( IOException x )
          {
            x.printStackTrace();
            setVisible( true );
          }
      }
    else if( o == cancel )
      {
        shutdown();
        System.exit( 1 );
      }
    else if( o == projectSignX )
      {
        if( projectSignX.isSelected() ||
            (projectSignL.isSelected() &&
             projectSignL.isEnabled()) )
          {
            projectSignInfoFile.setEnabled( true );
            projectSignInfoPick.setEnabled( true );
          }
        else
          {
            projectSignInfoFile.setEnabled( false );
            projectSignInfoPick.setEnabled( false );
          }
      }
    else if( o == projectSignL )
      {
        if( projectSignX.isSelected() ||
            (projectSignL.isSelected() &&
             projectSignL.isEnabled()) )
          {
            projectSignInfoFile.setEnabled( true );
            projectSignInfoPick.setEnabled( true );
          }
        else
          {
            projectSignInfoFile.setEnabled( false );
            projectSignInfoPick.setEnabled( false );
          }
      }
  }

  // ======================================================================
  // WindowListener
  // ======================================================================
  public void windowClosing( final WindowEvent e )
  {
    setVisible( false );
    System.exit( 1 );
  }
  public void windowClosed( final WindowEvent e )
  {
  }
  public void windowOpened( final WindowEvent e )
  {
    projectName.requestFocusInWindow();
  }
  public void windowIconified( final WindowEvent e )
  {
  }
  public void windowDeiconified( final WindowEvent e )
  {
  }
  public void windowActivated( final WindowEvent e )
  {
  }
  public void windowDeactivated( final WindowEvent e )
  {
  }

  private JTextField projectName;
  private JCheckBox projectWAR;
  private JTextField projectFile;
  private JTextField projectMain;
  private JTextField projectSVer;
  private JTextField projectTVer;
  private JRadioButton projectRepackageLibs;
  private JRadioButton projectReferenceLibs;
  private JRadioButton noLibraries;
  private JCheckBox projectImg;
  private JCheckBox projectSrc;
  private JCheckBox projectEtc;
  private JCheckBox projectHelp;
  private JCheckBox projectChkSum;
  private JCheckBox projectSignX;
  private JCheckBox projectSignL;
  private JTextField projectSignInfoFile;
  private JButton projectSignInfoPick;
  private JRadioButton projectGit;
  private JRadioButton projectSvn;
  private JRadioButton noLogs;
  private JComboBox<String> projectSkeleton;
  private JCheckBox projectJAXB;
  private JTextField pathSrc;
  private JTextField pathLib;
  private JTextField pathImg;
  private JTextField pathHelp;
  private JTextField pathEtc;
  private JTextField pathBuild;

  private JButton accept;
  private JButton cancel;

  private final String[] args;
  private static final String SUCCESS_MESSAGE =
    "+-----------------------------------------------------------------------+\n"+
    "| Successfully (re)created the file _build.xml (note the underscore)    |\n"+
    "| To use it as the default you need to copy/move/rename it to build.xml |\n"+
    "| You may want to test it first with a command like 'ant -f _build.xml' |\n"+
    "+-----------------------------------------------------------------------+\n";
  private static final Preferences prefs = Preferences.userNodeForPackage( Main.class );
}
